@@ -44,7 +44,9 @@ extern "C" unsigned int __builtin_get_sub_group_local_id();
 
 namespace sycl {
 __SYCL_INLINE_VER_NAMESPACE(_V1) {
-template <typename T, access::address_space Space> class multi_ptr;
+template <typename T, access::address_space Space,
+          access::decorated DecorateAddress>
+class multi_ptr;
 
 namespace detail {
 
@@ -65,11 +67,12 @@ using AcceptableForLocalLoadStore =
                   Space == access::address_space::local_space>;
 
 #ifdef __SYCL_DEVICE_ONLY__
-template <typename T, access::address_space Space>
-T load(const multi_ptr<T, Space> src) {
+template <typename T, access::address_space Space,
+          access::decorated DecorateAddress>
+T load(const multi_ptr<T, Space, DecorateAddress> src) {
   using BlockT = SelectBlockT<T>;
-  using PtrT =
-      sycl::detail::ConvertToOpenCLType_t<const multi_ptr<BlockT, Space>>;
+  using PtrT = sycl::detail::ConvertToOpenCLType_t<
+      const multi_ptr<BlockT, Space, DecorateAddress>>;
 
   BlockT Ret =
       __spirv_SubgroupBlockReadINTEL<BlockT>(reinterpret_cast<PtrT>(src.get()));
@@ -77,12 +80,13 @@ T load(const multi_ptr<T, Space> src) {
   return sycl::bit_cast<T>(Ret);
 }
 
-template <int N, typename T, access::address_space Space>
-vec<T, N> load(const multi_ptr<T, Space> src) {
+template <int N, typename T, access::address_space Space,
+          access::decorated DecorateAddress>
+vec<T, N> load(const multi_ptr<T, Space, DecorateAddress> src) {
   using BlockT = SelectBlockT<T>;
   using VecT = sycl::detail::ConvertToOpenCLType_t<vec<BlockT, N>>;
-  using PtrT =
-      sycl::detail::ConvertToOpenCLType_t<const multi_ptr<BlockT, Space>>;
+  using PtrT = sycl::detail::ConvertToOpenCLType_t<
+      const multi_ptr<BlockT, Space, DecorateAddress>>;
 
   VecT Ret =
       __spirv_SubgroupBlockReadINTEL<VecT>(reinterpret_cast<PtrT>(src.get()));
@@ -90,21 +94,24 @@ vec<T, N> load(const multi_ptr<T, Space> src) {
   return sycl::bit_cast<typename vec<T, N>::vector_t>(Ret);
 }
 
-template <typename T, access::address_space Space>
-void store(multi_ptr<T, Space> dst, const T &x) {
+template <typename T, access::address_space Space,
+          access::decorated DecorateAddress>
+void store(multi_ptr<T, Space, DecorateAddress> dst, const T &x) {
   using BlockT = SelectBlockT<T>;
-  using PtrT = sycl::detail::ConvertToOpenCLType_t<multi_ptr<BlockT, Space>>;
+  using PtrT = sycl::detail::ConvertToOpenCLType_t<
+      multi_ptr<BlockT, Space, DecorateAddress>>;
 
   __spirv_SubgroupBlockWriteINTEL(reinterpret_cast<PtrT>(dst.get()),
                                   sycl::bit_cast<BlockT>(x));
 }
 
-template <int N, typename T, access::address_space Space>
-void store(multi_ptr<T, Space> dst, const vec<T, N> &x) {
+template <int N, typename T, access::address_space Space,
+          access::decorated DecorateAddress>
+void store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, N> &x) {
   using BlockT = SelectBlockT<T>;
   using VecT = sycl::detail::ConvertToOpenCLType_t<vec<BlockT, N>>;
-  using PtrT =
-      sycl::detail::ConvertToOpenCLType_t<const multi_ptr<BlockT, Space>>;
+  using PtrT = sycl::detail::ConvertToOpenCLType_t<
+      const multi_ptr<BlockT, Space, DecorateAddress>>;
 
   __spirv_SubgroupBlockWriteINTEL(reinterpret_cast<PtrT>(dst.get()),
                                   sycl::bit_cast<VecT>(x));
@@ -113,10 +120,26 @@ void store(multi_ptr<T, Space> dst, const vec<T, N> &x) {
 
 } // namespace sub_group
 
+// Helper for removing const and volatile qualifiers from the element type of
+// a multi_ptr.
+template <typename CVT, access::address_space Space,
+          access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+inline multi_ptr<T, Space, IsDecorated>
+GetUnqualMultiPtr(const multi_ptr<CVT, Space, IsDecorated> &Mptr) {
+  if constexpr (IsDecorated == access::decorated::legacy) {
+    return multi_ptr<T, Space, IsDecorated>{
+        const_cast<typename multi_ptr<T, Space, IsDecorated>::pointer_t>(
+            Mptr.get())};
+  } else {
+    return multi_ptr<T, Space, IsDecorated>{
+        const_cast<typename multi_ptr<T, Space, IsDecorated>::pointer>(
+            Mptr.get_decorated())};
+  }
+}
+
 } // namespace detail
 
-namespace ext {
-namespace oneapi {
+namespace ext::oneapi {
 
 struct sub_group;
 namespace experimental {
@@ -203,8 +226,7 @@ struct sub_group {
 
   template <typename T>
   using EnableIfIsScalarArithmetic =
-      sycl::detail::enable_if_t<sycl::detail::is_scalar_arithmetic<T>::value,
-                                T>;
+      std::enable_if_t<sycl::detail::is_scalar_arithmetic<T>::value, T>;
 
   /* --- one-input shuffles --- */
   /* indices in [0 , sub_group size) */
@@ -258,19 +280,17 @@ struct sub_group {
 #ifdef __SYCL_DEVICE_ONLY__
   // Method for decorated pointer
   template <typename CVT, typename T = std::remove_cv_t<CVT>>
-  detail::enable_if_t<
-      !std::is_same<typename detail::remove_AS<T>::type, T>::value, T>
+  std::enable_if_t<!std::is_same<remove_decoration_t<T>, T>::value, T>
   load(CVT *cv_src) const {
     T *src = const_cast<T *>(cv_src);
-    return load(sycl::multi_ptr<typename detail::remove_AS<T>::type,
-                                sycl::detail::deduce_AS<T>::value>(
-        (typename detail::remove_AS<T>::type *)src));
+    return load(sycl::multi_ptr<remove_decoration_t<T>,
+                                sycl::detail::deduce_AS<T>::value,
+                                sycl::access::decorated::yes>(src));
   }
 
   // Method for raw pointer
   template <typename CVT, typename T = std::remove_cv_t<CVT>>
-  detail::enable_if_t<
-      std::is_same<typename detail::remove_AS<T>::type, T>::value, T>
+  std::enable_if_t<std::is_same<remove_decoration_t<T>, T>::value, T>
   load(CVT *cv_src) const {
     T *src = const_cast<T *>(cv_src);
 
@@ -299,11 +319,12 @@ struct sub_group {
 #endif //__SYCL_DEVICE_ONLY__
 
   template <typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value, T>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
 #ifdef __SYCL_DEVICE_ONLY__
 #ifdef __NVPTX__
     return src.get()[get_local_id()[0]];
@@ -318,11 +339,12 @@ struct sub_group {
   }
 
   template <typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForLocalLoadStore<T, Space>::value, T>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
 #ifdef __SYCL_DEVICE_ONLY__
     return src.get()[get_local_id()[0]];
 #else
@@ -334,12 +356,13 @@ struct sub_group {
 #ifdef __SYCL_DEVICE_ONLY__
 #ifdef __NVPTX__
   template <int N, typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value,
       vec<T, N>>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
     vec<T, N> res;
     for (int i = 0; i < N; ++i) {
       res[i] = *(src.get() + i * get_max_local_range()[0] + get_local_id()[0]);
@@ -348,60 +371,64 @@ struct sub_group {
   }
 #else  // __NVPTX__
   template <int N, typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
           N != 1 && N != 3 && N != 16,
       vec<T, N>>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
     return sycl::detail::sub_group::load<N, T>(src);
   }
 
   template <int N, typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
           N == 16,
       vec<T, 16>>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
     return {sycl::detail::sub_group::load<8, T>(src),
             sycl::detail::sub_group::load<8, T>(src +
                                                 8 * get_max_local_range()[0])};
   }
 
   template <int N, typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
           N == 3,
       vec<T, 3>>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
     return {
         sycl::detail::sub_group::load<1, T>(src),
         sycl::detail::sub_group::load<2, T>(src + get_max_local_range()[0])};
   }
 
   template <int N, typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
           N == 1,
       vec<T, 1>>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
     return sycl::detail::sub_group::load(src);
   }
 #endif // ___NVPTX___
 #else  // __SYCL_DEVICE_ONLY__
   template <int N, typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value,
       vec<T, N>>
-  load(const multi_ptr<CVT, Space> src) const {
+  load(const multi_ptr<CVT, Space, IsDecorated> src) const {
     (void)src;
     throw runtime_error("Sub-groups are not supported on host device.",
                         PI_ERROR_INVALID_DEVICE);
@@ -409,12 +436,13 @@ struct sub_group {
 #endif // __SYCL_DEVICE_ONLY__
 
   template <int N, typename CVT, access::address_space Space,
-            typename T = std::remove_cv_t<CVT>>
-  sycl::detail::enable_if_t<
+            access::decorated IsDecorated, typename T = std::remove_cv_t<CVT>>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForLocalLoadStore<T, Space>::value,
       vec<T, N>>
-  load(const multi_ptr<CVT, Space> cv_src) const {
-    multi_ptr<T, Space> src = const_cast<T *>(static_cast<CVT *>(cv_src));
+  load(const multi_ptr<CVT, Space, IsDecorated> cv_src) const {
+    multi_ptr<T, Space, IsDecorated> src =
+        sycl::detail::GetUnqualMultiPtr(cv_src);
 #ifdef __SYCL_DEVICE_ONLY__
     vec<T, N> res;
     for (int i = 0; i < N; ++i) {
@@ -431,20 +459,18 @@ struct sub_group {
 #ifdef __SYCL_DEVICE_ONLY__
   // Method for decorated pointer
   template <typename T>
-  detail::enable_if_t<
-      !std::is_same<typename detail::remove_AS<T>::type, T>::value>
-  store(T *dst, const typename detail::remove_AS<T>::type &x) const {
-    store(sycl::multi_ptr<typename detail::remove_AS<T>::type,
-                          sycl::detail::deduce_AS<T>::value>(
-              (typename detail::remove_AS<T>::type *)dst),
+  std::enable_if_t<!std::is_same<remove_decoration_t<T>, T>::value>
+  store(T *dst, const remove_decoration_t<T> &x) const {
+    store(sycl::multi_ptr<remove_decoration_t<T>,
+                          sycl::detail::deduce_AS<T>::value,
+                          sycl::access::decorated::yes>(dst),
           x);
   }
 
   // Method for raw pointer
   template <typename T>
-  detail::enable_if_t<
-      std::is_same<typename detail::remove_AS<T>::type, T>::value>
-  store(T *dst, const typename detail::remove_AS<T>::type &x) const {
+  std::enable_if_t<std::is_same<remove_decoration_t<T>, T>::value>
+  store(T *dst, const remove_decoration_t<T> &x) const {
 
 #ifdef __NVPTX__
     dst[get_local_id()[0]] = x;
@@ -475,10 +501,11 @@ struct sub_group {
   }
 #endif //__SYCL_DEVICE_ONLY__
 
-  template <typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value>
-  store(multi_ptr<T, Space> dst, const T &x) const {
+  store(multi_ptr<T, Space, DecorateAddress> dst, const T &x) const {
 #ifdef __SYCL_DEVICE_ONLY__
 #ifdef __NVPTX__
     dst.get()[get_local_id()[0]] = x;
@@ -493,10 +520,11 @@ struct sub_group {
 #endif
   }
 
-  template <typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForLocalLoadStore<T, Space>::value>
-  store(multi_ptr<T, Space> dst, const T &x) const {
+  store(multi_ptr<T, Space, DecorateAddress> dst, const T &x) const {
 #ifdef __SYCL_DEVICE_ONLY__
     dst.get()[get_local_id()[0]] = x;
 #else
@@ -509,55 +537,63 @@ struct sub_group {
 
 #ifdef __SYCL_DEVICE_ONLY__
 #ifdef __NVPTX__
-  template <int N, typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <int N, typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value>
-  store(multi_ptr<T, Space> dst, const vec<T, N> &x) const {
+  store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, N> &x) const {
     for (int i = 0; i < N; ++i) {
       *(dst.get() + i * get_max_local_range()[0] + get_local_id()[0]) = x[i];
     }
   }
 #else // __NVPTX__
-  template <int N, typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <int N, typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
       N != 1 && N != 3 && N != 16>
-  store(multi_ptr<T, Space> dst, const vec<T, N> &x) const {
+  store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, N> &x) const {
     sycl::detail::sub_group::store(dst, x);
   }
 
-  template <int N, typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <int N, typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
       N == 1>
-  store(multi_ptr<T, Space> dst, const vec<T, 1> &x) const {
+  store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, 1> &x) const {
     sycl::detail::sub_group::store(dst, x);
   }
 
-  template <int N, typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <int N, typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
       N == 3>
-  store(multi_ptr<T, Space> dst, const vec<T, 3> &x) const {
-    store<1, T, Space>(dst, x.s0());
-    store<2, T, Space>(dst + get_max_local_range()[0], {x.s1(), x.s2()});
+  store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, 3> &x) const {
+    store<1, T, Space, DecorateAddress>(dst, x.s0());
+    store<2, T, Space, DecorateAddress>(dst + get_max_local_range()[0],
+                                        {x.s1(), x.s2()});
   }
 
-  template <int N, typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <int N, typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value &&
       N == 16>
-  store(multi_ptr<T, Space> dst, const vec<T, 16> &x) const {
-    store<8, T, Space>(dst, x.lo());
-    store<8, T, Space>(dst + 8 * get_max_local_range()[0], x.hi());
+  store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, 16> &x) const {
+    store<8, T, Space, DecorateAddress>(dst, x.lo());
+    store<8, T, Space, DecorateAddress>(dst + 8 * get_max_local_range()[0],
+                                        x.hi());
   }
 
 #endif // __NVPTX__
 #else  // __SYCL_DEVICE_ONLY__
-  template <int N, typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <int N, typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForGlobalLoadStore<T, Space>::value>
-  store(multi_ptr<T, Space> dst, const vec<T, N> &x) const {
+  store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, N> &x) const {
     (void)dst;
     (void)x;
     throw runtime_error("Sub-groups are not supported on host device.",
@@ -565,10 +601,11 @@ struct sub_group {
   }
 #endif // __SYCL_DEVICE_ONLY__
 
-  template <int N, typename T, access::address_space Space>
-  sycl::detail::enable_if_t<
+  template <int N, typename T, access::address_space Space,
+            access::decorated DecorateAddress>
+  std::enable_if_t<
       sycl::detail::sub_group::AcceptableForLocalLoadStore<T, Space>::value>
-  store(multi_ptr<T, Space> dst, const vec<T, N> &x) const {
+  store(multi_ptr<T, Space, DecorateAddress> dst, const vec<T, N> &x) const {
 #ifdef __SYCL_DEVICE_ONLY__
     for (int i = 0; i < N; ++i) {
       *(dst.get() + i * get_max_local_range()[0] + get_local_id()[0]) = x[i];
@@ -775,8 +812,7 @@ inline sub_group this_sub_group() {
 #endif
 }
 
-} // namespace oneapi
-} // namespace ext
+} // namespace ext::oneapi
 
 } // __SYCL_INLINE_VER_NAMESPACE(_V1)
 } // namespace sycl
